/**
 * Register one to many tasks type to the WorkerTaskManager. Then init and enqueue a worker based execution by passing
 * configuration and buffers. The WorkerTaskManager allows to execute a maximum number of executions in parallel.
 *
 * Initial idea by Don McCurdy / https://www.donmccurdy.com / https://github.com/mrdoob/three.js/issues/18234
 */
declare class WorkerTaskManager {
    private taskTypes;
    private verbose;
    private maxParallelExecutions;
    private actualExecutionCount;
    private storedExecutions;
    /**
     * Creates a new WorkerTaskManager instance.
     *
     * @param {number} [maxParallelExecutions] How many workers are allowed to be executed in parallel.
     */
    constructor(maxParallelExecutions: number);
    /**
     * Set if logging should be verbose
     *
     * @param {boolean} verbose
     * @return {WorkerTaskManager}
     */
    setVerbose(verbose: boolean): this;
    /**
     * Set the maximum number of parallel executions.
     *
     * @param {number} maxParallelExecutions How many workers are allowed to be executed in parallel.
     * @return {WorkerTaskManager}
     */
    setMaxParallelExecutions(maxParallelExecutions: number): this;
    /**
     * Returns the maximum number of parallel executions.
     * @return {number}
     */
    getMaxParallelExecutions(): number;
    /**
     * Returns true if support for the given task type is available.
     *
     * @param {string} taskTypeName The task type as string
     * @return boolean
     */
    supportsTaskType(taskTypeName: string): boolean;
    /**
     * Registers functionality for a new task type based on module file.
     *
     * @param {string} taskTypeName The name to be used for registration.
     * @param {boolean} moduleWorker If the worker is a module or a standard worker
     * @param {URL} workerUrl The URL to be used for the Worker. Worker must provide logic to handle "init" and "execute" messages.
     * @return {boolean} Tells if registration is possible (new=true) or if task was already registered (existing=false)
     */
    registerTask(taskTypeName: string, moduleWorker: boolean, workerUrl: URL): boolean;
    /**
     * Provides initialization configuration and transferable objects.
     *
     * @param {string} taskTypeName The name of the registered task type.
     * @param {PayloadType} payload Configuration properties as serializable string.
     * @param {Transferable[]} [transferables] Any optional {@link ArrayBuffer} encapsulated in object.
     */
    initTaskType(taskTypeName: string, payload: PayloadType, transferables?: Transferable[]): Promise<void>;
    private wait;
    /**
     * Queues a new task of the given type. Task will not execute until initialization completes.
     *
     * @param {string} taskTypeName The name of the registered task type.
     * @param {PayloadType} payload Configuration properties as serializable string.
     * @param {(data: unknown) => void} assetAvailableFunction Invoke this function if an asset become intermediately available
     * @param {Transferable[]} [transferables] Any optional {@link ArrayBuffer} encapsulated in object.
     * @return {Promise}
     */
    enqueueForExecution(taskTypeName: string, payload: PayloadType, assetAvailableFunction?: (data: PayloadType) => void, transferables?: Transferable[]): Promise<unknown>;
    private depleteExecutions;
    /**
     * Destroys all workers and associated resources.
     * @return {WorkerTaskManager}
     */
    dispose(): this;
}
declare type PayloadType = {
    cmd: string;
    id: number;
    name: string | 'unknown';
    type?: string | 'unknown';
    workerId?: number;
    params?: Record<string, unknown>;
};
interface WorkerTaskManagerWorker {
    init(payload: PayloadType): void;
    execute(payload: PayloadType): void;
    comRouting(message: never): void;
}
declare class WorkerTaskManagerDefaultWorker implements WorkerTaskManagerWorker {
    init(payload: PayloadType): void;
    execute(payload: PayloadType): void;
    comRouting(message: MessageEvent<any>): void;
}
export { WorkerTaskManager, PayloadType, WorkerTaskManagerWorker, WorkerTaskManagerDefaultWorker };
//# sourceMappingURL=WorkerTaskManager.d.ts.map